@@ -23,40 +23,45 @@ import threading
 import types
 import uuid
 
+import netaddr
 from neutron_lib.api.definitions import portbindings
 from neutron_lib.api.definitions import provider_net
 from neutron_lib.api.definitions import segment as segment_def
-from neutron_lib.api.definitions import stateful_security_group
 from neutron_lib.callbacks import events
 from neutron_lib.callbacks import registry
 from neutron_lib.callbacks import resources
 from neutron_lib import constants as const
 from neutron_lib import context as n_context
+from neutron_lib.db import api as db_api
 from neutron_lib import exceptions as n_exc
 from neutron_lib.exceptions import availability_zone as az_exc
 from neutron_lib.placement import utils as place_utils
 from neutron_lib.plugins import directory
 from neutron_lib.plugins.ml2 import api
 from neutron_lib.utils import helpers
-from oslo_concurrency import lockutils
 from oslo_config import cfg
 from oslo_db import exception as os_db_exc
 from oslo_log import log
 from oslo_service import service as oslo_service
 from oslo_utils import timeutils
+from oslo_utils import uuidutils
 from ovsdbapp.backend.ovs_idl import idlutils
 
 from neutron._i18n import _
+from neutron.api import wsgi
 from neutron.common.ovn import acl as ovn_acl
 from neutron.common.ovn import constants as ovn_const
 from neutron.common.ovn import exceptions as ovn_exceptions
 from neutron.common.ovn import extensions as ovn_extensions
 from neutron.common.ovn import utils as ovn_utils
+from neutron.common import utils as n_utils
+from neutron.common import wsgi_utils
 from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
 from neutron.db import ovn_hash_ring_db
 from neutron.db import ovn_revision_numbers_db
 from neutron.db import provisioning_blocks
 from neutron.extensions import securitygroup as ext_sg
+from neutron.objects import router
 from neutron.plugins.ml2 import db as ml2_db
 from neutron.plugins.ml2.drivers.ovn.agent import neutron_agent as n_agent
 from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions \
@@ -71,7 +76,6 @@ from neutron.services.logapi.drivers.ovn import driver as log_driver
 from neutron.services.qos.drivers.ovn import driver as qos_driver
 from neutron.services.segments import db as segment_service_db
 from neutron.services.trunk.drivers.ovn import trunk_driver
-import neutron.wsgi
 
 
 LOG = log.getLogger(__name__)
@@ -122,7 +126,7 @@ class OVNMechanismDriver(api.MechanismDriver):
         self._maintenance_thread = None
         self._hash_ring_thread = None
         self._hash_ring_probe_event = multiprocessing.Event()
-        self.node_uuid = None
+        self._node_uuid = None
         self.hash_ring_group = ovn_const.HASH_RING_ML2_GROUP
         self.sg_enabled = ovn_acl.is_sg_enabled()
         ovn_conf.register_opts()
@@ -138,14 +142,12 @@ class OVNMechanismDriver(api.MechanismDriver):
                          OVN_MIN_GENEVE_MAX_HEADER_SIZE)
             raise SystemExit(1)
         self._setup_vif_port_bindings()
-        if impl_idl_ovn.OvsdbSbOvnIdl.schema_has_table('Chassis_Private'):
-            self.agent_chassis_table = 'Chassis_Private'
-        else:
-            self.agent_chassis_table = 'Chassis'
         self.subscribe()
         self.qos_driver = qos_driver.OVNQosDriver.create(self)
         self.trunk_driver = trunk_driver.OVNTrunkDriver.create(self)
         self.log_driver = log_driver.register(self)
+        self._start_time = None
+        self._agent_cache = None
 
     @property
     def nb_schema_helper(self):
@@ -186,6 +188,46 @@ class OVNMechanismDriver(api.MechanismDriver):
     def sb_ovn(self, val):
         self._sb_ovn = val
 
+    @property
+    def start_time(self):
+        if self._start_time:
+            return self._start_time
+
+        self._start_time = wsgi_utils.get_start_time()
+        if not self._start_time:
+            LOG.warning('uWSGI must provide a start time using the '
+                        'configuration parameter "start-time %t" in the '
+                        'configuration file')
+            # NOTE(ralonsoh): this is happening if the uWSGI configuration file
+            # does not have the "start-time %t" parameter or when using the
+            # Neutron API eventlet server, still in use in the grenade
+            # skip-level jobs. This should be removed in the F release.
+            self._start_time = wsgi_utils.get_start_time(current_time=True)
+
+        return self._start_time
+
+    @property
+    def node_uuid(self):
+        if self._node_uuid:
+            return self._node_uuid
+
+        worker_id = wsgi_utils.get_api_worker_id()
+        if worker_id is None:
+            # NOTE(ralonsoh): the hash ring node UUID should be based on the
+            # Neutron API worker ID. Right now only uWSGI mode is supported.
+            # The worker ID is provided via ``uwsgi`` library. If other loader
+            # is used, a random node UUID will be provided.
+            LOG.warning('uWSGI is the only supported loader for the Neutron '
+                        'API; it provides, via ``uwsgi`` library, the worker '
+                        'ID. If other loader is used, a random hash ring node '
+                        'UUID will be provided')
+            self._node_uuid = uuidutils.generate_uuid()
+        else:
+            self._node_uuid = ovn_hash_ring_db.get_node_uuid(
+                self.hash_ring_group, cfg.CONF.host, worker_id)
+
+        return self._node_uuid
+
     def get_supported_vif_types(self):
         vif_types = set()
         for ch in self.sb_ovn.chassis_list().execute(check_error=True):
@@ -208,6 +250,14 @@ class OVNMechanismDriver(api.MechanismDriver):
         return (context.current.get(provider_net.NETWORK_TYPE)
                 in vlan_transparency_network_types)
 
+    def check_vlan_qinq(self, context):
+        """OVN driver vlan QinQ support."""
+        vlan_qinq_network_types = [
+            const.TYPE_VLAN,
+        ]
+        return (context.current.get(provider_net.NETWORK_TYPE)
+                in vlan_qinq_network_types)
+
     def _setup_vif_port_bindings(self):
         self.supported_vnic_types = ovn_const.OVN_SUPPORTED_VNIC_TYPES
         self.vif_details = {
@@ -215,6 +265,10 @@ class OVNMechanismDriver(api.MechanismDriver):
                 portbindings.CAP_PORT_FILTER: self.sg_enabled,
                 portbindings.VIF_DETAILS_CONNECTIVITY: self.connectivity,
             },
+            portbindings.VIF_TYPE_AGILIO_OVS: {
+                portbindings.CAP_PORT_FILTER: self.sg_enabled,
+                portbindings.VIF_DETAILS_CONNECTIVITY: self.connectivity,
+            },
             portbindings.VIF_TYPE_VHOST_USER: {
                 portbindings.CAP_PORT_FILTER: False,
                 portbindings.VHOST_USER_MODE:
@@ -229,10 +283,7 @@ class OVNMechanismDriver(api.MechanismDriver):
         return portbindings.CONNECTIVITY_L2
 
     def supported_extensions(self, extensions):
-        supported_extensions = set(ovn_extensions.ML2_SUPPORTED_API_EXTENSIONS)
-        if not cfg.CONF.ovn.allow_stateless_action_supported:
-            supported_extensions.discard(stateful_security_group.ALIAS)
-        return set(supported_extensions) & extensions
+        return set(ovn_extensions.ML2_SUPPORTED_API_EXTENSIONS) & extensions
 
     @staticmethod
     def provider_network_attribute_updates_supported():
@@ -255,8 +306,13 @@ class OVNMechanismDriver(api.MechanismDriver):
         registry.subscribe(self.delete_segment_provnet_port,
                            resources.SEGMENT,
                            events.AFTER_DELETE)
+        # TODO(slaweq): use constant from neutron_lib.callbacks.resources once
+        # it will be available and released
+        registry.subscribe(self._validate_allowed_address_pairs,
+                           'allowed_address_pair',
+                           events.BEFORE_CREATE)
 
-        # Handle security group/rule notifications
+        # Handle security group/rule or address group notifications
         if self.sg_enabled:
             registry.subscribe(self._create_security_group_precommit,
                                resources.SECURITY_GROUP,
@@ -282,6 +338,15 @@ class OVNMechanismDriver(api.MechanismDriver):
             registry.subscribe(self._process_sg_rule_notification,
                                resources.SECURITY_GROUP_RULE,
                                events.BEFORE_DELETE)
+            registry.subscribe(self._process_ag_notification,
+                               resources.ADDRESS_GROUP,
+                               events.AFTER_CREATE)
+            registry.subscribe(self._process_ag_notification,
+                               resources.ADDRESS_GROUP,
+                               events.AFTER_UPDATE)
+            registry.subscribe(self._process_ag_notification,
+                               resources.ADDRESS_GROUP,
+                               events.AFTER_DELETE)
 
     def _remove_node_from_hash_ring(self, *args, **kwargs):
         # The node_uuid attribute will be empty for worker types
@@ -298,41 +363,52 @@ class OVNMechanismDriver(api.MechanismDriver):
 
     @staticmethod
     def should_post_fork_initialize(worker_class):
-        return worker_class in (neutron.wsgi.WorkerService,
+        return worker_class in (wsgi.WorkerService,
                                 worker.MaintenanceWorker,
                                 service.RpcWorker)
 
-    @lockutils.synchronized('hash_ring_probe_lock', external=True)
     def _setup_hash_ring(self):
         """Setup the hash ring.
 
-        The first worker to acquire the lock is responsible for cleaning
-        the hash ring from previous runs as well as start the probing
-        thread for this host. Subsequently workers just need to register
-        themselves to the hash ring.
+        The first worker to execute this method will remove the hash ring from
+        previous runs as well as start the probing thread for this host.
+        Subsequently workers just need to register themselves to the hash ring.
         """
         # Attempt to remove the node from the ring when the worker stops
         sh = oslo_service.SignalHandler()
         atexit.register(self._remove_node_from_hash_ring)
         sh.add_handler("SIGTERM", self._remove_node_from_hash_ring)
+        self._init_hash_ring(n_context.get_admin_context())
+        self._register_hash_ring_maintenance()
 
-        admin_context = n_context.get_admin_context()
-        if not self._hash_ring_probe_event.is_set():
-            # Clear existing entries
-            ovn_hash_ring_db.remove_nodes_from_host(admin_context,
-                                                    self.hash_ring_group)
-            self.node_uuid = ovn_hash_ring_db.add_node(admin_context,
-                                                       self.hash_ring_group)
-            self._hash_ring_thread = maintenance.MaintenanceThread()
-            self._hash_ring_thread.add_periodics(
-                maintenance.HashRingHealthCheckPeriodics(
-                    self.hash_ring_group))
-            self._hash_ring_thread.start()
-            LOG.info("Hash Ring probing thread has started")
-            self._hash_ring_probe_event.set()
-        else:
-            self.node_uuid = ovn_hash_ring_db.add_node(admin_context,
-                                                       self.hash_ring_group)
+    def _register_hash_ring_maintenance(self):
+        """Maintenance method for the node OVN hash ring register
+
+        The ``self.node_uuid`` value must be set before calling this method.
+        """
+        self._hash_ring_thread = maintenance.MaintenanceThread()
+        self._hash_ring_thread.add_periodics(
+            maintenance.HashRingHealthCheckPeriodics(
+                self.hash_ring_group, self.node_uuid))
+        self._hash_ring_thread.start()
+        LOG.info('Hash Ring probing thread for node %s has started',
+                 self.node_uuid)
+
+    @db_api.retry_if_session_inactive()
+    @db_api.CONTEXT_WRITER
+    def _init_hash_ring(self, context):
+        LOG.debug('Hash Ring setup using WSGI start time %s',
+                  str(n_utils.ts_to_datetime(self.start_time)))
+        created_at = n_utils.ts_to_datetime(self.start_time)
+        ovn_hash_ring_db.remove_nodes_from_host(
+            context, self.hash_ring_group, created_at=created_at)
+        ovn_hash_ring_db.add_node(
+            context, self.hash_ring_group, self.node_uuid,
+            created_at=created_at)
+        newer_nodes = ovn_hash_ring_db.get_nodes(
+            context, self.hash_ring_group, created_at=created_at)
+        LOG.debug('Hash Ring setup, this worker has detected %s OVN hash '
+                  'ring registers in the database', len(newer_nodes))
 
     def post_fork_initialize(self, resource, event, trigger, payload=None):
         # Initialize API/Maintenance workers with OVN IDL connections
@@ -343,10 +419,11 @@ class OVNMechanismDriver(api.MechanismDriver):
         self._post_fork_event.clear()
         self._ovn_client_inst = None
 
-        if worker_class == neutron.wsgi.WorkerService:
+        if worker_class == wsgi.WorkerService:
             self._setup_hash_ring()
 
-        n_agent.AgentCache(self)  # Initialize singleton agent cache
+        # Initialize singleton agent cache and keep a copy.
+        self._agent_cache = n_agent.AgentCache(self)
         self.nb_ovn, self.sb_ovn = impl_idl_ovn.get_ovn_idls(self, trigger)
 
         # Override agents API methods
@@ -426,21 +503,16 @@ class OVNMechanismDriver(api.MechanismDriver):
     def _delete_security_group(self, resource, event, trigger, payload):
         context = payload.context
         security_group_id = payload.resource_id
-        self._ovn_client.delete_security_group(context,
-                                               security_group_id)
+        self._ovn_client.delete_security_group(
+            context, security_group_id, delete_sg_rules=True)
 
     def _update_security_group(self, resource, event, trigger, payload):
         context = payload.context
         security_group = payload.latest_state
 
         old_state, new_state = payload.states
-        is_allow_stateless_supported = (
-            self._ovn_client.is_allow_stateless_supported()
-        )
-        old_stateful = ovn_acl.is_sg_stateful(
-            old_state, is_allow_stateless_supported)
-        new_stateful = ovn_acl.is_sg_stateful(
-            new_state, is_allow_stateless_supported)
+        old_stateful = ovn_acl.is_sg_stateful(old_state)
+        new_stateful = ovn_acl.is_sg_stateful(new_state)
         if old_stateful != new_stateful:
             for rule in self._plugin.get_security_group_rules(
                     context, {'security_group_id': [security_group['id']]}):
@@ -501,6 +573,25 @@ class OVNMechanismDriver(api.MechanismDriver):
                 return True
         return False
 
+    def _process_ag_notification(
+            self, resource, event, trigger, payload):
+        context = payload.context
+        address_group = payload.latest_state
+        address_group_id = payload.resource_id
+        if event == events.AFTER_CREATE:
+            ovn_revision_numbers_db.create_initial_revision(
+                context, address_group_id, ovn_const.TYPE_ADDRESS_GROUPS,
+                std_attr_id=address_group['standard_attr_id'])
+            self._ovn_client.create_address_group(
+                context, address_group)
+        elif event == events.AFTER_UPDATE:
+            self._ovn_client.update_address_group(
+                context, address_group)
+        elif event == events.AFTER_DELETE:
+            self._ovn_client.delete_address_group(
+                context,
+                address_group_id)
+
     def _is_network_type_supported(self, network_type):
         return (network_type in [const.TYPE_LOCAL,
                                  const.TYPE_FLAT,
@@ -538,6 +629,57 @@ class OVNMechanismDriver(api.MechanismDriver):
                 )
                 raise n_exc.InvalidInput(error_message=m)
 
+    def _validate_allowed_address_pairs(self, resource, event, trigger,
+                                        payload):
+        context = payload.desired_state['context']
+        allowed_address_pairs = payload.desired_state['allowed_address_pairs']
+        network_id = payload.desired_state['network_id']
+        if not allowed_address_pairs:
+            return
+
+        port_allowed_address_pairs_ip_addresses = [
+            netaddr.IPNetwork(pair['ip_address'])
+            for pair in allowed_address_pairs]
+
+        distributed_ports = self._plugin.get_ports(
+            context.elevated(),
+            filters={'device_owner': [const.DEVICE_OWNER_DISTRIBUTED],
+                     'network_id': [network_id]})
+        if not distributed_ports:
+            return
+
+        def _get_common_ips(ip_addresses, ip_networks):
+            common_ips = set()
+            for ip_address in ip_addresses:
+                if any(ip_address in ip_net for ip_net in ip_networks):
+                    common_ips.add(str(ip_address))
+            return common_ips
+
+        # NOTE(slaweq): We can safely ignore any CIDR larger than /32 (for
+        # IPv4) or /128 (for IPv6) in the allowed_address_pairs, since such
+        # CIDRs cannot be set as a Virtual IP in OVN.
+        # Only /32 and /128 CIDRs are allowed to be set as Virtual IPs in OVN.
+        address_pairs_to_check = [
+            ip_net for ip_net in port_allowed_address_pairs_ip_addresses
+            if ip_net.size == 1]
+
+        for distributed_port in distributed_ports:
+            distributed_port_ip_addresses = [
+                netaddr.IPAddress(fixed_ip['ip_address']) for fixed_ip in
+                distributed_port.get('fixed_ips', [])]
+
+            common_ips = _get_common_ips(
+                distributed_port_ip_addresses,
+                address_pairs_to_check)
+
+            if common_ips:
+                err_msg = (
+                    _("IP addresses '%s' already used by the '%s' port(s) in "
+                      "the same network" % (";".join(common_ips),
+                                            const.DEVICE_OWNER_DISTRIBUTED))
+                )
+                raise n_exc.InvalidInput(error_message=err_msg)
+
     def create_segment_provnet_port(self, resource, event, trigger,
                                     payload=None):
         segment = payload.latest_state
@@ -695,12 +837,18 @@ class OVNMechanismDriver(api.MechanismDriver):
 
         # in the case of router ports we also need to
         # track the creation and update of the LRP OVN objects
-        if ovn_utils.is_lsp_router_port(port):
+        if (ovn_utils.is_lsp_router_port(port) and
+                self._is_ovn_router_flavor_port(context, port)):
             ovn_revision_numbers_db.create_initial_revision(
                 context.plugin_context, port['id'],
                 ovn_const.TYPE_ROUTER_PORTS,
                 std_attr_id=context.current['standard_attr_id'])
 
+    def _is_ovn_router_flavor_port(self, context, port):
+        router_obj = router.Router.get_object(context.plugin_context,
+                                              id=port['device_id'])
+        return ovn_utils.is_ovn_provider_router(router_obj)
+
     def _is_port_provisioning_required(self, port, host, original_host=None):
         vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
         if vnic_type not in self.supported_vnic_types:
@@ -838,7 +986,8 @@ class OVNMechanismDriver(api.MechanismDriver):
             self._insert_port_provisioning_block(context.plugin_context,
                                                  port['id'])
 
-        if ovn_utils.is_lsp_router_port(port):
+        if (ovn_utils.is_lsp_router_port(port) and
+                self._is_ovn_router_flavor_port(context, port)):
             # handle the case when an existing port is added to a
             # logical router so we need to track the creation of the lrp
             if not ovn_utils.is_lsp_router_port(original_port):
@@ -872,11 +1021,11 @@ class OVNMechanismDriver(api.MechanismDriver):
         # the port status from DOWN to UP in order to generate 'fake'
         # vif-interface-plugged event. This workaround is needed to
         # perform live-migration with live_migration_wait_for_vif_plug=True.
-        if ((port['status'] == const.PORT_STATUS_DOWN and
-             ovn_const.MIGRATING_ATTR in port[portbindings.PROFILE].keys() and
-             port[portbindings.VIF_TYPE] in (
+        if (port['status'] == const.PORT_STATUS_DOWN and
+            ovn_const.MIGRATING_ATTR in port[portbindings.PROFILE].keys() and
+            port[portbindings.VIF_TYPE] in (
                  portbindings.VIF_TYPE_OVS,
-                 portbindings.VIF_TYPE_VHOST_USER))):
+                 portbindings.VIF_TYPE_VHOST_USER)):
             LOG.info("Setting port %s status from DOWN to UP in order "
                      "to emit vif-interface-plugged event.",
                      port['id'])
@@ -1026,7 +1175,7 @@ class OVNMechanismDriver(api.MechanismDriver):
                          {'port_id': port['id'],
                           'network_type': network_type})
 
-            if ((network_type in ['flat', 'vlan']) and
+            if ((network_type in [const.TYPE_FLAT, const.TYPE_VLAN]) and
                     (physical_network not in chassis_physnets)):
                 LOG.info('Refusing to bind port %(port_id)s on '
                          'host %(host)s due to the OVN chassis '
@@ -1048,17 +1197,23 @@ class OVNMechanismDriver(api.MechanismDriver):
                     vif_details = copy.deepcopy(self.vif_details[vif_type])
                     vif_details[portbindings.VHOST_USER_SOCKET] = (
                         vhost_user_socket)
+                elif (vnic_type == portbindings.VNIC_VIRTIO_FORWARDER):
+                    vhost_user_socket = ovn_utils.ovn_vhu_sockpath(
+                        ovn_conf.get_ovn_vhost_sock_dir(), port['id'])
+                    vif_type = portbindings.VIF_TYPE_AGILIO_OVS
+                    port[portbindings.VIF_DETAILS].update({
+                        portbindings.VHOST_USER_SOCKET: vhost_user_socket})
+                    vif_details = copy.deepcopy(self.vif_details[vif_type])
+                    vif_details[portbindings.VHOST_USER_SOCKET] = (
+                        vhost_user_socket)
+                    vif_details[portbindings.VHOST_USER_MODE] = (
+                        portbindings.VHOST_USER_MODE_CLIENT)
                 else:
                     vif_type = portbindings.VIF_TYPE_OVS
                     vif_details = copy.deepcopy(self.vif_details[vif_type])
 
-                if self.agent_chassis_table == 'Chassis_Private':
-                    chassis_to_retrieve = agent.chassis_private
-                else:
-                    chassis_to_retrieve = agent.chassis
-                ovn_bridge = ovn_utils.get_ovn_bridge_from_chassis(
-                    chassis_to_retrieve)
-
+                ovn_bridge = ovn_utils.get_ovn_bridge_from_chassis_private(
+                    agent.chassis_private)
                 dp_type = ovn_utils.get_datapath_type(bind_host, self.sb_ovn)
                 vif_details.update({
                     portbindings.VIF_DETAILS_BRIDGE_NAME: ovn_bridge,
@@ -1195,9 +1350,9 @@ class OVNMechanismDriver(api.MechanismDriver):
                 self.sb_ovn.get_extport_chassis_from_cms_options()):
             try:
                 with self.nb_ovn.transaction(check_error=True) as txn:
-                    ovn_utils.sync_ha_chassis_group(
-                        admin_context, db_port['id'], db_port['network_id'],
-                        self.nb_ovn, self.sb_ovn, txn)
+                    ovn_utils.sync_ha_chassis_group_network(
+                        admin_context, self.nb_ovn, self.sb_ovn,
+                        db_port['id'], db_port['network_id'], txn)
             except Exception as e:
                 LOG.error('Error while syncing the HA Chassis Group for the '
                           'external port %s during set port status up. '
@@ -1223,7 +1378,7 @@ class OVNMechanismDriver(api.MechanismDriver):
 
             if self._should_notify_nova(db_port):
                 self._plugin.nova_notifier.record_port_status_changed(
-                    db_port, const.PORT_STATUS_ACTIVE, const.PORT_STATUS_DOWN,
+                    db_port, const.PORT_STATUS_DOWN, const.PORT_STATUS_ACTIVE,
                     None)
                 self._plugin.nova_notifier.send_port_status(
                     None, None, db_port)
@@ -1259,7 +1414,7 @@ class OVNMechanismDriver(api.MechanismDriver):
 
         available_seg_ids = {
             segment['id'] for segment in segments
-            if segment['network_type'] in ('flat', 'vlan')}
+            if segment['network_type'] in (const.TYPE_FLAT, const.TYPE_VLAN)}
 
         segment_service_db.update_segment_host_mapping(
             ctx, host, available_seg_ids)
@@ -1398,7 +1553,7 @@ class OVNMechanismDriver(api.MechanismDriver):
                 LOG.debug('Chassis %s is reponsible of the resource provider '
                           '%s', ch_name, ch_rp)
                 return True
-            elif len(states) > 1:
+            if len(states) > 1:
                 rps = {state[0]: placement_ext.dict_chassis_config(state[1])
                        for state in states}
                 LOG.error('Several chassis reported the requested resource '
@@ -1417,7 +1572,7 @@ def get_agents(self, context, filters=None, fields=None, _driver=None):
 
 def get_agent(self, context, id, fields=None, _driver=None):
     try:
-        return n_agent.AgentCache()[id].as_dict()
+        return n_agent.AgentCache().get(id).as_dict()
     except KeyError:
         raise n_exc.agent.AgentNotFound(id=id)
 
@@ -1431,17 +1586,13 @@ def update_agent(self, context, id, agent, _driver=None):
     # and we can just fall through to raising in the case that admin_state_up
     # is being set to False, otherwise the end-state will be fine
     if not agent.get('admin_state_up', True):
-        pass
-    elif 'description' in agent:
+        raise n_exc.BadRequest(resource='agent',
+                               msg='OVN agent status cannot be updated')
+    if 'description' in agent:
         _driver.sb_ovn.set_chassis_neutron_description(
             chassis_name, agent['description'],
             agent_type).execute(check_error=True)
-        return agent
-    else:
-        # admin_state_up=True w/o description
-        return agent
-    raise n_exc.BadRequest(resource='agent',
-                           msg='OVN agent status cannot be updated')
+    return agent
 
 
 def delete_agent(self, context, id, _driver=None):
